#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: julien de Rosny (c) Institut Langevin/ESPCI/CNRS
"""
from numpy import *
from scipy.optimize import minimize
from matplotlib.pyplot import *
from mutual import *
import matplotlib.pyplot as plt
import mpl_toolkits.mplot3d as plt3d

class PlotWires:
    def __init__(self):
        fig = plt.figure()
        fig.set_size_inches(10,10)
        self.ax = fig.add_subplot(111, projection='3d')
        self.ax.view_init(120,30)
    
    def plot(self,wires,colorLine='gray'):
        #plot the nodes
        for segment in wires:
            for i in range(2):
                self.ax.scatter(segment[i,0], segment[i,1], segment[i,2], color='black', marker='s')
            self.ax.plot3D(segment[:,0],segment[:,1],segment[:,2], colorLine)



close('all')

N = 8 # Number of Legs
widthLeg = 0.01 # Width of the leg
lengthLeg = 24e-2 # Length of legs
widthRing = widthLeg # Width of rings
R0 = 15.5e-2   # Radius of shield
Rleg = 13e-2 # Radius of circle where the leg are positionned
omega = 300e6*2*pi # Working frequency
deltaThetaLengthRing = 0.005/Rleg

# Definition of a function to transform a 2M vector into a vector of M+1 value,
# is just the complex sum of the M first elements
# This is done because we know that it is not possible to generate constant c
# current with a metacage
def Complexi(x):
    v=x[0::2]+1j*x[1::2]
    somme=np.sum(v)
    return np.append(v,-somme)

theta = np.linspace(0,2*np.pi*(1-1/N),N)
deltaTheta = theta[1] - theta[0]
# xy are the positions of the legs (required to compute mutual impedance) 
xyz  = Rleg *  column_stack((cos(theta),sin(theta),zeros((len(theta))))).transpose()
# Radius of the circle of the image of the leg circle with respect to shield
Rp = R0**2/Rleg
# xyp positions of the leg images
xyzp  = Rp *  column_stack((cos(theta),sin(theta),zeros((len(theta))))).transpose()
#
# xy are the positions of the legs (required to compute mutual impedance) 
ringSegments  = [\
                 array([[Rleg*cos(t+deltaThetaLengthRing)           ,Rleg*sin(t+deltaThetaLengthRing)           ,0], \
                        [Rleg*cos(t+deltaTheta-deltaThetaLengthRing),Rleg*sin(t+deltaTheta-deltaThetaLengthRing),0]]) \
                 for t in theta\
                ]
                 
legSegments = [ array([[Rleg*cos(t),Rleg*sin(t),0],[Rleg*cos(t),Rleg*sin(t),lengthLeg]])  for t in theta ]

plt3 = PlotWires()
plt3.plot(ringSegments)
plt3.plot([s + array([[0,0,lengthLeg],[0,0,lengthLeg]]) for s in ringSegments])
plt3.plot(legSegments)
show()
#

# Calculus of the self and mutual inductance
LLleg =  zeros((N,N))
for i in range(N):
    for j in range(N):
        dima =  norm([xyz[:,i]-xyzp[:,j]]) 
        if i!=j:
            d =  norm([xyz[:,i]-xyz[:,j]]) 
            LLleg[i,j] = Lnm(lengthLeg,d) - Lnm(lengthLeg,dima)
        else:
            LLleg[i,j] =  selfInducRect(lengthLeg,widthRing)-Lnm(lengthLeg,dima)#selfInducTub(lengthLeg,a1,a2)-Lnm(lengthLeg,dima)
        
LLring = zeros((N,N))
for i in range(N):
    for j in range(N):
        #print(i,j)

        if i!=j:
            
#            LLring[i,j] = MNonPara( column_stack((xyz[:,i],xyz[:,(i+1)%len(theta)])).transpose(), \
#                                    column_stack((xyz[:,j],xyz[:,(j+1)%len(theta)])).transpose())
            LLring[i,j] = MNonPara( ringSegments[i],ringSegments[j])
            if 0:
                clf()
                plot([xyz[0,i],xy[0,(i+1)%len(theta)]],[xyz[1,i],xyz[1,(i+1)%len(theta)]])
                plot([xyz[0,j],xy[0,(j+1)%len(theta)]],[xyz[1,j],xyz[1,(j+1)%len(theta)]])
                pause(.1)
        else:
            l = norm(xyz[:,i]-xyz[:,(i+1)%len(theta)])
            LLring[i,j] = selfInducRect(l,widthRing)
            #LLring[i,j] -= MNonPara(column_stack((xyz[:,i],xyz[:,(i+1)%len(theta)])).transpose(), \
            #                     column_stack((xyzp[:,j],xyzp[:,(j+1)%len(theta)])).transpose())




# Current distribution in case of a BIRDCAGE
Il0 = exp(1j*theta)

Il = Il0 - mean(Il0)

# Now from this current distribution that is compatible with metacage, we need 
# to deduce the inductance and capacitance
# To that end we introduce a matrix mat (see document)
CurrentToImpedMat = zeros((N,2*N),dtype=np.complex)
for i in range(N):
    CurrentToImpedMat[i,(2*(i+1))%(2*N)]  = Il[(i+1)%N]
    CurrentToImpedMat[i,(2*i+1)%(2*N)]  = -Il[i] 
    CurrentToImpedMat[i,2*i%(2*N)] = -Il[i%N]
    CurrentToImpedMat[i,(2*(i-1)+1)%(2*N)]  = Il[(i-1)%N]


#NEW PART
LegToRing = zeros((N,N))
for i in range(N):
    LegToRing[i,i]  = 1
    LegToRing[i,(i+1)%(N)]  = -1

umat,smat,vhmat = linalg.svd(LegToRing)

sinv = zeros(smat.shape)
sinv[0:-1]=1/smat[0:-1]
# The pseudo inverse is computed
invmat = vhmat.transpose() @ (sinv[...,None] *umat.transpose())

Ir = invmat @ Il

# We write equations in matricial form
# The input vector is [... gamma^L_n gamma^R_n gamma^L_(n+1) gamma^R_(n+1) ...]
# Its dimension is 2N
# The outpout is real part of the equation and imagingary ...
equs = zeros((2*N,2*N))
u = zeros((2*N))
Zmr = zeros((N),dtype=complex)
Zml = zeros((N),dtype=complex)

for i in range(N):
    for j in range(N):
        if 1:#i!=j:
            Zml[i] += LLleg[i,j]*Il[j]/Il[i]*1e-7*1j*omega/2.0 #Only the half of the leg impedance is taken into account (because the ground is at the middle)
            Zmr[i] += LLring[i,j]*Ir[j]/Ir[i]*1e-7*1j*omega

for i in range(N):
    equs[2*i   ,2*(i-1) + 0] = - Il[i-1].imag
    equs[2*i   ,2* i    + 0] = + Il[ i ].imag
    equs[2*i   ,2* i    + 1] = + Ir[ i ].imag
    u[2*i] = - Ir[i].imag*Zmr[i].imag - Il[i].imag*Zml[i].imag
    u[2*i] +=  Ir[i].real*Zmr[i].real + Il[i].real*Zml[i].real
    u[2*i] +=  -Il[i-1].real*Zml[i-1].real + Il[i-1].imag*Zml[i-1].imag

    equs[2*i + 1,2*(i-1) + 0] = + Il[i-1].real
    equs[2*i + 1,2* i    + 0] = - Il[i  ].real
    equs[2*i + 1,2* i    + 1] = - Ir[ i ].real
    u[2*i + 1 ] =   Ir[i].imag*Zmr[i].real + Il[i].real*Zml[i].imag
    u[2*i + 1 ] +=  Ir[i].real*Zmr[i].imag + Il[i].imag*Zml[i].real
    u[2*i + 1 ] +=  -Il[i-1].real*Zml[i-1].imag - Il[i-1].imag*Zml[i-1].real
    

umat,smat,vhmat = linalg.svd(equs)

sinv = zeros(smat.shape)
sinv[0:-1]=1/smat[0:-1]
invmat = vhmat.transpose() @ (sinv[...,None] *umat.transpose())
# One solution is computed
v0 = invmat@u
# Kernel vector is extracted
vker = vhmat[-1,:]

# All the impedance choice are OK to kernel vector.
# if ind_imped = 0, the choice is to have zero lumped impedance on the Legs
# if ind_imped = 1, the choice is to have zero lumped impedance on the Ringd
ind_imped = 1
choice =  v0 - v0[ind_imped]*vker/vker[ind_imped]

imZl=choice[0::2] 
imZr=choice[1::2] 

print('Rings')
#we compute self or capacitance
for i,z in enumerate(imZr):
    if z>0:
        print("Lr_%d = %e H"%(i,z/omega))
    else:
        print("Cr_%d = %e F"%(i,-1/omega/z))
        
print('Legs')

#we compute self or capacitance ()
for i,z in enumerate(imZl):
    if z>0:
        print("Ll_%d = %e H"%(i,z/omega))
    else:
        print("Cl_%d = %e F"%(i,-1/omega/z))

print('------------------')


ZtildeRing=fft.fft(LLring[:,0]*omega*1e-7)/sqrt(N)
ZtildeLeg=fft.fft(LLleg[:,0]*omega*1e-7)/sqrt(N)
zetaR= -sqrt(N)*(ZtildeRing+sin(pi/N*arange(N))/N*ZtildeLeg)

CR=-1/zetaR/omega    
CR[1]


